// Soubor SEZNAM.H - ablona jednosmrnho seznamu se zarkou
//
// Obsahuje metody pro vkldn na potek a na konec seznamu,
// vyprzdnn seznamu, zruen seznamu, pstup k hodnot uloen v prvnm prvku,
// vyjmut prvnho prvku seznamu a zjitn, zda je seznam przdn.
//
// V ppad nespchu vyvolvaj nkter metody vjimky typu chyba_seznamu:
// Jde o tyto metody: konstruktor, VlozNaPocatek a VlozNaKonec, pokud nen pam
//     pro alokovan prvek
// VyjmiPrvn a obsahPrvniho, pokud je seznam przdn
//
// Seznam nelze koprovat. Toho lze doshnout nsledujcm trikem:
// V tle tdy seznam deklarujeme koprovac konstruktor a piazovac
// opertor jako soukrom metody. Jejich tla nedefinujeme, nejsou teba.
// Peklada ozna pokus o koprovn jako chybu, protoe tyto metody nejsou
// dostupn.
// Na potku souboru me bt #definovno makro LADIM. Je-li definovno, stane se soust
// tohoto souboru funkce main, ve kter lze seznam testovat.

#define LADIM		
#include <iostream>
#include <stdexcept>

using namespace std;

class chyba_seznamu: public out_of_range {
public:
	chyba_seznamu(const string &s): out_of_range(s) {}
};

template<typename T> class prvek {
	T _hodnota;
	prvek * _dalsi;
public:
	prvek(T hodnota=0): _hodnota(hodnota), _dalsi(0) {}
	T &hodnota() { return _hodnota; }
	prvek *& dalsi() { return _dalsi; }
};

template<typename T> class seznam {
	prvek<T> *hlava, *zarazka;
	seznam(seznam&);			// Zakazujeme koprovn
	seznam& operator=(seznam&);
public:
	seznam() throw(chyba_seznamu);	// Konstruktor vytvo przdn seznam
	~seznam() throw();					// Destruktor vyprzdn seznam a zru i zarku
	void Vyprazdni() throw();			// Odstran ze seznamu vechny prvky krom zarky
	void VlozNaPocatek(T data) throw(chyba_seznamu);
	void VlozNaKonec(T data) throw(chyba_seznamu);
	T& ObsahPrvniho()  throw(chyba_seznamu);
	T VyjmiPrvni()  throw(chyba_seznamu);
	bool Prazdny() throw() { return hlava == zarazka; }
        prvek<T> * Najdi(T d) throw();
};

template<typename T>
seznam<T>::seznam() throw(chyba_seznamu): hlava(0), zarazka(0){	// Nekodn inicializace
        try {
        	hlava = new prvek<T>;						// Alokace zarky
	        zarazka = hlava;						// V przdnm seznamu ukazuje hlava i zarazka na t prvek - zarku
        }
        catch(bad_alloc)
        {
                throw chyba_seznamu("nepodarilo se vytvorit seznam");
        }
}


template<typename T>
seznam<T>::~seznam() throw() {
	Vyprazdni();			// Vyprzdni seznam
	delete hlava;			// Odstra zarku
}

template<typename T>		// Odstran vechny prvky krom zarky
void seznam<T>::Vyprazdni() throw(){
	while(hlava != zarazka)
	{
		prvek<T>* p = hlava;	// Ulo ukazatel na prvn prvek
		hlava = p->dalsi();	// Do ukazatele na prvn prvek ulo adresu druhho
		delete p;		// Vyma prvek, kter byl prvn
	}
}

template<typename T>
void seznam<T>::VlozNaPocatek(T d) throw(chyba_seznamu) {
	try{ 
		prvek<T> *p = new prvek<T>(d);	// Alokuj nov prvek
		p->dalsi() = hlava;		// Jeho ukazatel na nsledujc bude ukazovat na prvn prvek seznamu	
		hlava = p;			// Ukazatel na prvn prvek seznamu ukazuje na nov prvek
	}
	catch(bad_alloc)
	{
		throw chyba_seznamu("neni pamet pro novy prvek");
	}
}

template<typename T>
T& seznam<T>::ObsahPrvniho() throw (chyba_seznamu)
{
	if(Prazdny())
		throw chyba_seznamu("nelze menit prvni prvek prazdneho seznamu");
	else
		return hlava->hodnota();
}

template<typename T>
void seznam<T>::VlozNaKonec(T d) throw(chyba_seznamu)
{
	try {
		prvek<T>* p = new prvek<T>(0);	// Alokuj nov prvek
		zarazka->hodnota() = d;			// Ulo data do zarky
		zarazka->dalsi() = p;			// Nov prvek pipoj jako zarku
		zarazka = p;
	}
	catch(bad_alloc)
	{
		throw chyba_seznamu("neni pamet ");
	}
}

template<typename T>
T seznam<T>::VyjmiPrvni() throw (chyba_seznamu)
{
	if(Prazdny()) throw chyba_seznamu("Vyber prvniho prvku z prazdneho seznamu");
	prvek<T> *p = hlava;	// Zapamatuj si ukazatel na prvn prvek seznamu
	T d = p -> hodnota();	// Zapamatuj si data z prvnho prvku
	hlava = p->dalsi();	// Ukazatel na prvn prvek bude ukazovat na druh
	delete p;		// Vyma bval prvn prvek seznamu
	return d;		// Vra zapamatovanou hodnotu
}

template<typename T>
prvek<T> * seznam<T>::Najdi(T d) throw()
{
   zarazka -> hodnota() = d;     // Ulo hledanou hodnotu do zarky (take je jist, e v seznamu je)
   prvek<T>* p = hlava;          // Pomocn ukazatel na aktuln prvek seznamu - zanme u prvnho prvku
   while(p->hodnota() != d)      // Pokud p neukazuje na prvek obsahujc d, jdi na dal
      p = p->dalsi();
   return p == zarazka ? 0 : p;  // Ukazuje-li p na zarku, hledan hodnota v seznamu nen, vra NULL
}                                // Jinak vra adresu prvku

#ifdef LADIM
// Pomocn funkce main pro ladn seznamu
void main(){
	seznam<int> S;
    int i;
	for(i = 1; i < 5; i++)
		S.VlozNaPocatek(i);
	S.ObsahPrvniho() = 10;
	S.VlozNaKonec(12);
	i = S.VyjmiPrvni();
    prvek<int> *p = S.Najdi(3);
	S.Vyprazdni();
        // seznam<int> T(S); // Chyba - koprovac konstruktor je nepstupn
        seznam<int> U;
        //U = S;             // Chyba - piazovac opertor je nepstupn
}

#endif